/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mips.h"
#include "pic32.h"
#include "vga.h"

/* Disable JTAG functionality on pins. */

.section .devcfg0, "a"
.word 0xfffffffb	/* DEVCFG0<2> = JTAGEN = 0 */

/*
Set the oscillator to be the FRC oscillator with PLL, with peripheral clock
divided by 2, and FRCDIV+PLL selected.

The watchdog timer (FWDTEN) is also disabled.

The secondary oscillator pin (FSOSCEN) is disabled to avoid pin conflicts with
RPB4.
*/

.section .devcfg1, "a"
.word 0xff7fdfd9	/* DEVCFG1<23> = FWDTEN = 0; DEVCFG1<13:12> = FPBDIV<1:0> = 1;
			DEVCFG1<5> = FSOSCEN = 0; DEVCFG1<2:0> = FNOSC<2:0> = 001 */

/*
Set the FRC oscillator PLL function with an input division of 4, an output
division of 2, a multiplication of 24, yielding a multiplication of 3.

The FRC is apparently at 16MHz and this produces a system clock of 48MHz.
*/

.section .devcfg2, "a"
.word 0xfff9fffb	/* DEVCFG2<18:16> = FPLLODIV<2:0> = 001;
			DEVCFG2<6:4> = FPLLMUL<2:0> = 111;
			DEVCFG2<2:0> = FPLLIDIV<2:0> = 011 */

.text
.globl _start
.extern init_framebuffer
.extern init_framebuffer_with_pattern
.extern screendata
.extern fontdata
.extern blit_string
.extern message

_start:
	/*
	Configure RAM.
	See: http://microchipdeveloper.com/32bit:mx-arch-exceptions-processor-initialization
	*/

	la $v0, BMXCON
	lw $v1, 0($v0)

	/* Set zero wait states for address setup. */

	li $t8, ~(1 << 6)	/* BMXCON<6> = BMXWSDRM = 0 */
	and $v1, $v1, $t8

	/* Set bus arbitration mode. */

	li $t8, ~0b111		/* BMXCON<2:0> = BMXARB<2:0> = 0 */
	ori $t8, $t8, 0b010	/* BMXCON<2:0> = BMXARB<2:0> = 2 */
	and $v1, $v1, $t8
	sw $v1, 0($v0)

	/* Enable caching. */

	mfc0 $v1, CP0_CONFIG
	li $t8, ~CONFIG_K0
	and $v1, $v1, $t8
	ori $v1, $v1, CONFIG_K0_CACHABLE_NONCOHERENT
	mtc0 $v1, CP0_CONFIG
	nop

	/* Get the RAM size. */

	la $v0, BMXDRMSZ
	lw $t0, 0($v0)

	/* Initialise the stack pointer. */

	li $v1, KSEG0_BASE
	addu $sp, $t0, $v1	/* sp = KSEG0_BASE + RAM size */

	/* Initialise the globals pointer. */

	lui $gp, %hi(_GLOBAL_OFFSET_TABLE_)
	ori $gp, $gp, %lo(_GLOBAL_OFFSET_TABLE_)

	/* Set pins for output. */

	jal init_pins
	nop

	la $t0, PORTA
	li $t1, (1 << 3)	/* PORTA<3> = RA3 */
	sw $t1, CLR($t0)

	jal init_oc_pins
	nop

	/* Initialise the status register. */

	jal init_interrupts
	nop

	/* Initialise framebuffer. */

	la $a0, screendata
	jal init_framebuffer
	nop

	sync

	/* Initialise timer. */

	jal init_timer2
	nop

	/* Initialise DMA. */

	jal init_dma
	nop

	/* Initialise OC1 and OC2. */

	jal init_oc
	nop

	/* Initialise the display state. */

	li $s0, 0		/* line counter */
	la $s1, vbp_active	/* current event */
	li $s2, SCREEN_BASE	/* line address */
	li $s3, SCREEN_BASE	/* screen address */

	/* Save the state for retrieval in the interrupt handler. */

	li $k0, IRQ_STACK_LIMIT
	sw $s0, -44($k0)
	sw $s1, -48($k0)
	sw $s2, -52($k0)
	sw $s3, -56($k0)

	/* Enable interrupts and loop. */

	jal enable_interrupts
	nop

	jal handle_error_level
	nop

	/* Main program. */

	li $a1, (3 << 24)	/* counter ~= 50000000 */
	li $a2, 0xffffff	/* test counter at every 1/4 of range */
	move $t2, $zero		/* picture to show */

	/* Monitoring loop. */
loop:
	addiu $a1, $a1, -1	/* counter -= 1 */
	and $t1, $a2, $a1
	bnez $t1, loop
	nop

	la $t0, PORTA
	li $t1, (1 << 3)	/* PORTA<3> = RA3 */
	sw $t1, INV($t0)

	bnez $a1, loop		/* until counter == 0 */
	nop

	bnez $t2, _picture1
	nop

	/* Show picture 0. */

	la $a0, screendata
	jal init_framebuffer
	nop

	la $a0, message0
	li $a1, SCREEN_BASE_KSEG0
	jal blit_string
	nop

	li $t2, 1
	j _next
	nop

_picture1:
	/* Show picture 1. */

	jal init_framebuffer_with_pattern
	nop

	la $a0, message1
	li $a1, SCREEN_BASE_KSEG0
	jal blit_string
	nop

	move $t2, $zero

_next:
	li $a1, (3 << 24)	/* counter ~= 50000000 */
	li $a2, 0xffffff	/* test counter at every 1/4 of range */
	j loop
	nop



init_pins:
	/* DEVCFG0<2> needs setting to 0 before the program is run. */

	la $v0, CFGCON
	li $v1, (1 << 3)	/* CFGCON<3> = JTAGEN = 0 */
	sw $v1, CLR($v0)

init_outputs:
	/* Remove analogue features from pins. */

	la $v0, ANSELA
	sw $zero, 0($v0)	/* ANSELA = 0 */
	la $v0, ANSELB
	sw $zero, 0($v0)	/* ANSELB = 0 */

	la $v0, TRISA
	sw $zero, 0($v0)
	la $v0, TRISB
	sw $zero, 0($v0)

	la $v0, PORTA
	sw $zero, 0($v0)
	la $v0, PORTB
	sw $zero, 0($v0)

	jr $ra
	nop



/* Initialisation routines. */

init_timer2:

	/* Initialise Timer2 for sync pulses. */

	la $v0, T2CON
	sw $zero, 0($v0)	/* T2CON = 0 */
	nop

	la $v0, TMR2
	sw $zero, 0($v0)	/* TMR2 = 0 */

	la $v0, PR2
	li $v1, HFREQ_LIMIT
	sw $v1, 0($v0)		/* PR2 = HFREQ_LIMIT */

	/* Initialise Timer2 interrupt. */

	la $v0, IFS0
	li $v1, (1 << 9)
	sw $v1, CLR($v0)	/* T2IF = 0 */

	la $v0, IPC2
	li $v1, 0b11111
	sw $v1, CLR($v0)	/* T2IP, T2IS = 0 */
	li $v1, 0b11111
	sw $v1, SET($v0)	/* T2IP = 7; T2IS = 3 */

	la $v0, IEC0
	li $v1, (1 << 9)
	sw $v1, SET($v0)	/* T2IE = 1 */

	/* Start timer. */

	la $v0, T2CON
	li $v1, (1 << 15)
	sw $v1, SET($v0)	/* ON = 1 */

	jr $ra
	nop



/*
Output compare initialisation.

Timer2 will be used to trigger two events using OC1: one initiating the hsync
pulse, and one terminating the pulse. The pulse should appear after the line
data has been transferred using DMA, but this is achieved by just choosing
suitable start and end values.

Using OC2, Timer2 triggers a level shifting event and OC2 is reconfigured to
reverse the level at a later point. In this way, the vsync pulse is generated
and is synchronised to the display lines.
*/

init_oc:
	/* Disable OC1 interrupts. */

	la $v0, IEC0
	li $v1, (1 << 7)	/* IEC0<7> = OC1IE = 0 */
	sw $v1, CLR($v0)

	la $v0, IFS0
	li $v1, (1 << 7)	/* IFS0<7> = OC1IF = 0 */
	sw $v1, CLR($v0)

	/* Initialise OC1. */

	la $v0, OC1CON
	li $v1, 0b101		/* OC1CON<2:0> = OCM<2:0> = 101 (dual compare, continuous pulse) */
	sw $v1, 0($v0)

	/* Pulse start and end. */

	la $v0, OC1R
	li $v1, HSYNC_END	/* HSYNC_START for positive polarity */
	sw $v1, 0($v0)

	la $v0, OC1RS
	li $v1, HSYNC_START	/* HSYNC_END for positive polarity */
	sw $v1, 0($v0)

	/* OC1 is enabled. */

	la $v0, OC1CON
	li $v1, (1 << 15)
	sw $v1, SET($v0)

	/* Disable OC2 interrupts. */

	la $v0, IEC0
	li $v1, (1 << 12)	/* IEC0<12> = OC2IE = 0 */
	sw $v1, CLR($v0)

	la $v0, IFS0
	li $v1, (1 << 12)	/* IFS0<12> = OC2IF = 0 */
	sw $v1, CLR($v0)

	/* Initialise OC2. */

	la $v0, OC2CON
	li $v1, 0b010		/* OC2CON<2:0> = OCM<2:0> = 010 (single compare, output driven low) */
	sw $v1, 0($v0)

	/* Set pulse position. */

	la $v0, OC2R
	sw $zero, 0($v0)

	/* Enable OC2 later. */

	jr $ra
	nop

init_oc_pins:
	/* Unlock the configuration register bits. */

	la $v0, SYSKEY
	sw $zero, 0($v0)
	li $v1, 0xAA996655
	sw $v1, 0($v0)
	li $v1, 0x556699AA
	sw $v1, 0($v0)

	la $v0, CFGCON
	lw $t8, 0($v0)
	li $v1, (1 << 13)	/* IOLOCK = 0 */
	sw $v1, CLR($v0)

	/* Map OC1 to RPA0. */

	la $v0, RPA0R
	li $v1, 0b0101		/* RPA0R<3:0> = 0101 (OC1) */
	sw $v1, 0($v0)

	/* Map OC2 to RPA1. */

	la $v0, RPA1R
	li $v1, 0b0101		/* RPA1R<3:0> = 0101 (OC2) */
	sw $v1, 0($v0)

	la $v0, CFGCON
	sw $t8, 0($v0)

	/* Lock the oscillator control register again. */

	la $v0, SYSKEY
	li $v1, 0x33333333
	sw $v1, 0($v0)

	jr $ra
	nop



/*
Direct Memory Access initialisation.

Write 160 pixels to PORTB for the line data. This is initiated by a timer
interrupt. Upon completion of the transfer, a DMA interrupt initiates the
address update routine, changing the source address of the DMA channel.
*/

init_dma:
	/* Disable DMA interrupts. */

	la $v0, IEC1
	li $v1, (3 << 28)	/* IEC1<29:28> = DMA1IE, DMA0IE = 0 */
	sw $v1, CLR($v0)

	/* Clear DMA interrupt flags. */

	la $v0, IFS1
	li $v1, (3 << 28)	/* IFS1<29:28> = DMA1IF, DMA0IF = 0 */
	sw $v1, CLR($v0)

	/* Enable DMA. */

	la $v0, DMACON
	li $v1, (1 << 15)
	sw $v1, SET($v0)

	/*
	Initialise a line channel.
	The line channel will be channel 0 (x = 0).

	Specify a priority of 3:
	DCHxCON<1:0> = CHPRI<1:0> = 3

	Auto-enable the channel:
	DCHxCON<4> = CHAEN = 1
	*/

	la $v0, DCH0CON
	li $v1, 0b10011
	sw $v1, 0($v0)

	/*
	Initialise a level reset channel.
	The reset channel will be channel 1 (x = 1).

	Specify a priority of 3:
	DCHxCON<1:0> = CHPRI<1:0> = 3

	Chain the channel to channel 0:
	DCHxCON<5> = CHCHN = 1

	Allow the channel to receive events when disabled:
	DCHxCON<6> = CHAED = 1
	*/

	la $v0, DCH1CON
	li $v1, 0b1100011
	sw $v1, 0($v0)

	/*
	Initiate channel transfers when the initiating interrupt condition
	occurs:
	DCHxECON<15:8> = CHSIRQ<7:0> = timer 2 interrupt
	DCHxECON<4> = SIRQEN = 1

	For now, however, prevent initiation by not setting SIRQEN.
	*/

	la $v0, DCH0ECON
	li $v1, (9 << 8)
	sw $v1, 0($v0)

	/*
	Initiate reset channel transfer when channel 0 is finished:
	DCHxECON<15:8> = CHSIRQ<7:0> = channel 0 interrupt
	DCHxECON<4> = SIRQEN = 1
	*/

	la $v0, DCH1ECON
	li $v1, (60 << 8) | (1 << 4)
	sw $v1, 0($v0)

	/*
	The line channel has a cell size of the number bytes in a line:
	DCHxCSIZ<15:0> = CHCSIZ<15:0> = LINE_LENGTH
	*/

	la $v0, DCH0CSIZ
	li $v1, LINE_LENGTH
	sw $v1, 0($v0)

	/*
	The reset channel has a cell size of a single zero byte:
	DCHxCSIZ<15:0> = CHCSIZ<15:0> = 1
	*/

	la $v0, DCH1CSIZ
	li $v1, 1
	sw $v1, 0($v0)

	/*
	The source has a size identical to the cell size:
	DCHxSSIZ<15:0> = CHSSIZ<15:0> = LINE_LENGTH or 1
	*/

	la $v0, DCH0SSIZ
	li $v1, LINE_LENGTH
	sw $v1, 0($v0)

	la $v0, DCH1SSIZ
	li $v1, 1
	sw $v1, 0($v0)

	/*
	The source address is the physical address of the line data:
	DCHxSSA = physical(line data address)
	*/

	la $v0, DCH0SSA
	li $v1, SCREEN_BASE
	sw $v1, 0($v0)

	/*
	For the reset channel, a single byte of zero is transferred:
	DCHxSSA = physical(zero data address)
	*/

	la $v0, DCH1SSA
	la $v1, zerodata
	li $t8, KSEG0_BASE
	subu $v1, $v1, $t8
	sw $v1, 0($v0)

	/*
	The destination has a size of 1 byte:
	DCHxDSIZ<15:0> = CHDSIZ<15:0> = 1
	*/

	la $v0, DCH0DSIZ
	li $v1, 1
	sw $v1, 0($v0)

	la $v0, DCH1DSIZ
	sw $v1, 0($v0)

	/*
	The destination address is the physical address of PORTB:
	DCHxDSA = physical(PORTB)
	*/

	la $v0, DCH0DSA
	li $v1, PORTB
	li $t8, KSEG1_BASE
	subu $v1, $v1, $t8
	sw $v1, 0($v0)

	la $v0, DCH1DSA
	sw $v1, 0($v0)

	/*
	Use the block transfer completion interrupt to indicate when the source
	address can be updated.
	*/

	la $v0, DCH0INT
	li $v1, (1 << 19)	/* CHBCIE = 1 */
	sw $v1, 0($v0)

	/* Enable interrupt for address updating. */

	la $v0, IPC10
	li $v1, 0b11111		/* DMA0IP, DMA0IS = 0 */
	sw $v1, CLR($v0)

	la $v0, IPC10
	li $v1, 0b11111		/* DMA0IP = 7, DMA0IS = 3 */
	sw $v1, SET($v0)

	la $v0, IEC1
	li $v1, (1 << 28)	/* IEC1<28> = DMA0IE = 1 */
	sw $v1, SET($v0)

	/* Enable line channel. */

	la $v0, DCH0CON
	li $v1, 0b10000000
	sw $v1, SET($v0)

	jr $ra
	nop

zerodata:
.word 0



/* Utilities. */

handle_error_level:
	mfc0 $t3, CP0_STATUS
	li $t4, ~(STATUS_ERL | STATUS_EXL)
	and $t3, $t3, $t4
	mtc0 $t3, CP0_STATUS
	jr $ra
	nop

enable_interrupts:
	mfc0 $t3, CP0_STATUS
	li $t4, ~STATUS_IRQ	/* Clear interrupt priority bits. */
	and $t3, $t3, $t4
	ori $t3, $t3, (3 << STATUS_IRQ_SHIFT)
	li $t4, ~STATUS_BEV	/* CP0_STATUS &= ~STATUS_BEV (use non-bootloader vectors) */
	and $t3, $t3, $t4
	ori $t3, $t3, STATUS_IE
	mtc0 $t3, CP0_STATUS
	jr $ra
	nop

init_interrupts:
	mfc0 $t3, CP0_DEBUG
	li $t4, ~DEBUG_DM
	and $t3, $t3, $t4
	mtc0 $t3, CP0_DEBUG

	mfc0 $t3, CP0_STATUS
	li $t4, STATUS_BEV	/* BEV = 1 or EBASE cannot be set */
	or $t3, $t3, $t4
	mtc0 $t3, CP0_STATUS

	la $t3, exception_handler
	mtc0 $t3, CP0_EBASE	/* EBASE = exception_handler */

	li $t3, 0x20		/* Must be non-zero or the CPU gets upset */
	mtc0 $t3, CP0_INTCTL

	li $t3, CAUSE_IV	/* IV = 1 (use EBASE+0x200 for interrupts) */
	mtc0 $t3, CP0_CAUSE

	jr $ra
	nop



/* Exception servicing. */

.section .flash, "a"

/* TLB error servicing. */

tlb_handler:
	j exception_handler
	nop



/* General exception servicing. */

.org 0x180

exception_handler:
	j exc_handler
	nop



/* Interrupt servicing. */

.org 0x200

interrupt_handler:

	/* Store affected registers. */

	li $k0, IRQ_STACK_LIMIT
	sw $v0, -4($k0)
	sw $v1, -8($k0)
	sw $s0, -12($k0)
	sw $s1, -16($k0)
	sw $s2, -20($k0)
	sw $s3, -24($k0)
	sw $t8, -28($k0)
	sw $ra, -32($k0)
	sw $sp, -36($k0)

	/* Load state. */

	lw $s0, -44($k0)
	lw $s1, -48($k0)
	lw $s2, -52($k0)
	lw $s3, -56($k0)

	li $sp, IRQ_STACK_TOP

	/* Check for a timer interrupt condition. */

	la $v0, IFS0
	lw $v1, 0($v0)
	andi $v1, $v1, (1 << 9)	/* T2IF */
	beqz $v1, irq_dma
	nop

	/* Clear the timer interrupt condition. */

	sw $v1, CLR($v0)

	/*
	The timer interrupt will only occur outside the visible region, but the
	interrupt condition will still occur as the timer wraps around.
	Therefore, the handling of other interrupts may find the timer interrupt
	condition set.
	
	For the visible region, the event handler is invoked when handling the
	DMA interrupt. Otherwise, the event handler is invoked in response to
	the timer interrupt.
	*/

	la $t8, visible_active
	beq $s1, $t8, irq_dma
	nop

	/* Increment the line counter (only outside the visible region). */

	addiu $s0, $s0, 1

	/* Jump to the event handler (only outside the visible region). */

	jalr $s1
	nop

irq_dma:
	/* Check for a DMA interrupt condition. */

	la $v0, IFS1
	lw $v1, 0($v0)
	li $t8, (1 << 28)	/* DMA0IF */
	and $v1, $v1, $t8
	beqz $v1, irq_exit
	nop

	/* Clear the DMA interrupt condition. */

	sw $v1, CLR($v0)

	/* Test the block transfer completion interrupt flag. */

	la $v0, DCH0INT
	lw $v1, 0($v0)
	andi $v1, $v1, (1 << 3)	/* CHBCIF */
	beqz $v1, irq_exit
	nop

	/* Clear the block transfer completion interrupt flag. */

	sw $v1, CLR($v0)

	/*
	The DMA interrupt should only be active within the visible region.
	The event handler is invoked here instead of in response to a timer
	interrupt within that region.
	*/

	/* Increment the line counter (only within the visible region). */

	addiu $s0, $s0, 1

	/* Jump to the event handler (only within the visible region). */

	jalr $s1
	nop

	/* Jump to the DMA update routine. */

	j visible_update_address
	nop

irq_exit:
	/* Save state. */

	li $k0, IRQ_STACK_LIMIT
	sw $s0, -44($k0)
	sw $s1, -48($k0)
	sw $s2, -52($k0)
	sw $s3, -56($k0)

	/* Restore affected registers. */

	lw $v0, -4($k0)
	lw $v1, -8($k0)
	lw $s0, -12($k0)
	lw $s1, -16($k0)
	lw $s2, -20($k0)
	lw $s3, -24($k0)
	lw $t8, -28($k0)
	lw $ra, -32($k0)
	lw $sp, -36($k0)

	eret
	nop



exc_handler:
	li $t9, 0x80000000
	mfc0 $t6, CP0_ERROREPC
	nop
exc_loop:
	and $t7, $t9, $t6
	beqz $t7, exc_errorepc_zero
	nop
exc_errorepc_one:
	la $v0, PORTA
	li $v1, (1 << 2)	/* PORTA<2> = RA2 */
	sw $v1, SET($v0)
	j exc_loop_wait
	nop
exc_errorepc_zero:
	la $v0, PORTA
	li $v1, (1 << 3)	/* PORTA<3> = RA3 */
	sw $v1, SET($v0)
exc_loop_wait:
	li $t8, 5000000
exc_loop_delay:
	addiu $t8, $t8, -1
	bnez $t8, exc_loop_delay
	nop
	la $v0, PORTA
	li $v1, (3 << 2)	/* PORTA<3:2> = RA3, RA2 */
	sw $v1, CLR($v0)
exc_loop_wait_again:
	li $t8, 2500000
exc_loop_delay_again:
	addiu $t8, $t8, -1
	bnez $t8, exc_loop_delay_again
	nop
exc_errorepc_next:
	srl $t9, $t9, 1
	bnez $t9, exc_loop
	nop
	j exc_handler
	nop



/* Event routines. */

/* The vertical back porch. */

vbp_active:
	/* Test for visible region. */

	sltiu $v0, $s0, VISIBLE_START
	bnez $v0, _vbp_active_ret
	nop

	/* Start the visible region. */

	la $s1, visible_active

	/* Reset the line address. */

	move $s2, $s3

	/* Update the source address. */

	la $v0, DCH0SSA
	sw $s2, 0($v0)

	/* Enable the line channel for timer event transfer initiation. */

	la $v0, DCH0ECON
	li $v1, (1 << 4)	/* DCH0ECON<4> = SIRQEN = 1 */
	sw $v1, SET($v0)

	/*
	Suspend delivery of the timer interrupt during the visible period.
	The condition still occurs, however.
	*/

	la $v0, IEC0
	li $v1, (1 << 9)
	sw $v1, CLR($v0)	/* T2IE = 0 */

	la $v0, IPC2
	li $v1, 0b11111
	sw $v1, CLR($v0)	/* T2IP, T2IS = 0 */
	li $v1, 0b00111
	sw $v1, SET($v0)	/* T2IP = 1; T2IS = 3 */

	la $v0, IEC0
	li $v1, (1 << 9)
	sw $v1, SET($v0)	/* T2IE = 0 */

_vbp_active_ret:
	jr $ra
	nop



/* The visible region. */

visible_active:
	/* Test for front porch. */

	sltiu $v0, $s0, VFP_START
	bnez $v0, _visible_active_ret
	nop

	/* Start the front porch region. */

	la $s1, vfp_active

	/* Restore delivery of the timer interrupt after the visible period. */

	la $v0, IEC0
	li $v1, (1 << 9)
	sw $v1, CLR($v0)	/* T2IE = 0 */

	la $v0, IPC2
	li $v1, 0b11111
	sw $v1, CLR($v0)	/* T2IP, T2IS = 0 */
	li $v1, 0b11111
	sw $v1, SET($v0)	/* T2IP = 7; T2IS = 3 */

	la $v0, IEC0
	li $v1, (1 << 9)
	sw $v1, SET($v0)	/* T2IE = 1 */

_visible_active_ret:
	jr $ra
	nop



/* DMA update routine. */

visible_update_address:

	/* Test for the last visible line. */

	la $v0, vfp_active
	bne $s1, $v0, _visible_update_address
	nop

	/* Disable the line channel. */

	la $v0, DCH0ECON
	li $v1, (1 << 4)	/* DCH0ECON<4> = SIRQEN = 0 */
	sw $v1, CLR($v0)

	j _visible_update_ret
	nop

_visible_update_address:

	/*
	Update the line data address if the line counter (referring to the
	next line) is even.
	*/

	andi $t8, $s0, 1
	bnez $t8, _visible_update_ret
	nop

	/* Reference the next line and update the DMA source address. */

	addiu $s2, $s2, LINE_LENGTH

	/* Test for wraparound. */

	li $t8, (SCREEN_BASE + SCREEN_SIZE)
	sltu $t8, $s2, $t8
	bnez $t8, _visible_dma_update
	nop

	/* Reset the source address. */

	li $s2, SCREEN_BASE

_visible_dma_update:

	/* Update the source address. */

	la $v0, DCH0SSA
	sw $s2, 0($v0)

_visible_update_ret:
	j irq_exit
	nop



/* Within the vertical front porch. */

vfp_active:
	/* Test for vsync. */

	sltiu $v0, $s0, VSYNC_START
	bnez $v0, _vfp_active_ret
	nop

	/* Start the vsync. */

	la $s1, vsync_active

	/* Bring vsync low when the next line starts. */

	la $v0, OC2CON
	li $v1, 0b010 | (1 << 15)	/* OC2CON<2:0> = OCM<2:0> = 010 (single compare, output driven low) */
	sw $v1, 0($v0)

_vfp_active_ret:
	jr $ra
	nop



/* The vsync period. */

vsync_active:
	/* Test for front porch. */

	sltiu $v0, $s0, VSYNC_END
	bnez $v0, _vsync_active_ret
	nop

	/* Start the back porch. */

	move $s0, $zero
	la $s1, vbp_active

	/* Bring vsync high when the next line starts. */

	la $v0, OC2CON
	li $v1, 0b001 | (1 << 15)	/* OC2CON<2:0> = OCM<2:0> = 001 (single compare, output driven high) */
	sw $v1, 0($v0)

_vsync_active_ret:
	jr $ra
	nop
